from collections.abc import Iterable
from urllib.parse import quote

import pandas as pd
from sqlalchemy import create_engine, engine

from connectors_to_databases.BaseOperator import BaseOperator
from connectors_to_databases.helper_functions.function_list import (
    list_values_in_str_with_double_quotes,
    list_values_in_str_with_single_quotes,
)


class PostgreSQL(BaseOperator):
    """Connector to PostgreSQL database."""

    def __init__(
            self,
            host: str = "localhost",
            port: int = 5432,
            database: str = "postgres",
            login: str = "postgres",
            password: str = "postgres",  # noqa: S107
    ):
        """
        Init class.

        :param host: Host/IP database; default 'localhost'.
        :param database: name database; default 'localhost'.
        :param port: port database; default 5432.
        :param login: login to database; default 'postgres'.
        :param password: password to database; default 'postgres'.
        """
        super().__init__(host, port, database, login, password)
        self._host = host
        self._database = database
        self._login = login
        self._password = password
        self._port = port

    def _authorization_database(self) -> engine.base.Engine:
        """Creating connector engine to database PostgreSQL."""  # noqa: D401

        engine_str = f"postgresql://" \
                     f"{self._login}:{quote(self._password)}@{self._host}:{self._port}/" \
                     f"{self._database}"

        return create_engine(engine_str)

    @classmethod
    def generate_on_conflict_sql_query(
            cls,
            source_table_schema_name: str = "public",
            source_table_name: str = None,
            target_table_schema_name: str = "public",
            target_table_name: str = None,
            list_columns: Iterable[str] = None,
            pk: str | list[str, ...] = "id",
            replace: bool = False,
    ) -> str:
        """
        **Function: generate_on_conflict_sql_query.**

        This class method generates an SQL query for performing data insertion with conflict handling using a
        specified primary key.

        **Parameters:**
        - `source_table_schema_name` (str, optional): The schema name of the source table. Defaults to `'public'`.
        - `source_table_name` (str): The name of the source table from which data will be inserted.
        - `target_table_schema_name` (str, optional): The schema name of the target table where data will be inserted.
            Defaults to `'public'`.
        - `target_table_name` (str): The name of the target table where data will be inserted.
        - `list_columns` (Iterable[str], optional): An iterable object containing the names of columns to be inserted.
            If not specified, all columns from the source table will be inserted.
        - `pk` (str | list[str, ...]): The primary key for checking insertion conflicts. It can be a string representing
            a single column name or a list of column names.
        - `replace` (bool): A flag indicating whether to replace existing data in case of conflicts. Defaults to
            `False`, which means conflicts will be ignored and nothing will be done.

        **Return:**
        - `str`: The generated SQL query for data insertion with conflict handling.

        **Example Usage:**

        ```python
        source_table_name = 'source_table'
        target_table_name = 'target_table'
        columns = ['column1', 'column2', 'column3']
        sql_query = MyClass.generate_on_conflict_sql_query(
            source_table_name=source_table_name,
            target_table_name=target_table_name,
            list_columns=columns,
            pk='id',
            replace=True
        )
        print(sql_query)
        ```

        **Output:**

        ```
        INSERT INTO public.target_table
        (
            "column1", "column2", "column3"
        )
        SELECT
            "column1", "column2", "column3"
        FROM
            public.source_table
        ON CONFLICT ("id") DO UPDATE SET
            "column1" = EXCLUDED."column1",
            "column2" = EXCLUDED."column2",
            "column3" = EXCLUDED."column3"
        ```

        In this example, we pass the source table name (`source_table_name`), target table name (`target_table_name`),
        column list (`columns`), primary key (`pk`), and the `replace` flag to the `generate_on_conflict_sql_query`
        function. Then, we store the result in the `sql_query` variable and print it. The output will contain the
        generated SQL query for data insertion with conflict handling.

        :param source_table_schema_name: The schema name of the source table; default `'public'`.
        :param source_table_name: The name of the source table from which data will be inserted; default `None`.
        :param target_table_schema_name: The schema name of the target table where data will be
            inserted; default `'public'`.
        :param target_table_name: The name of the target table where data will be inserted; default `None`.
        :param list_columns: An iterable object containing the names of columns to be inserted; default `None`.
        :param pk: The primary key for checking insertion conflicts. It can be a string representing a single
        column name or a list of column names; default `'id'`.
        :param replace: A flag indicating whether to replace existing data in case of conflicts.
        Defaults to `False`, which means conflicts will be ignored and nothing will be done; default `False`.
        :return: The generated SQL query for data insertion with conflict handling.
        """ # noqa D415

        pk = list_values_in_str_with_double_quotes(list_columns=pk) if isinstance(pk, list) else f'"{pk}"'

        if replace:
            replace = f"""DO UPDATE SET {', '.join([f'"{i}" = EXCLUDED."{i}"' for i in list_columns])}"""
        else:
            replace = "DO NOTHING"

        sql = f"""
        INSERT INTO {target_table_schema_name}.{target_table_name}
        (
            {list_values_in_str_with_double_quotes(list_columns=list_columns)}
        )
        SELECT
            {list_values_in_str_with_double_quotes(list_columns=list_columns)}
        FROM
            {source_table_schema_name}.{source_table_name}
        ON CONFLICT ({pk}) {replace}
        """

        return sql  # noqa: RET504

    def get_database_description(
            self,
            table_name: str | list[str, ...] = None,
            table_schema: str | list[str, ...] = None,
    ) -> pd.DataFrame:
        """
        **Function: get_database_description.**

        This method retrieves descriptive information about tables, columns, and their descriptions within a
        PostgreSQL database schema.
        
        **Parameters:**
        - `table_name` (str | list[str, ...]): The name of the table or list of tables. If provided, the function 
        will filter results to include only the specified table(s). Default is `None`.
        - `table_schema` (str | list[str, ...]): The schema name of the table or list of tables. If provided, the        function will filter results to include only the specified schema(s). Default is `None`.
        
        **Return:**
        - `pd.DataFrame`: A pandas DataFrame containing descriptive information about the tables and columns.
        
        **Example Usage:**
        
        ```python
        # Instantiate an object of the class
        pg = PostgreSQL()
        
        # Retrieve description for a specific table
        table_description = pg.get_database_description(table_name="employees")
        
        # Retrieve description for tables in a schema
        schema_description = pg.get_database_description(table_schema="public")

        :param table_name: The table name.
        :param table_schema: The table schema.
        :return: pd.DataFrame with descriptive information about the tables and columns.
        """ # noqa D415

        where_condition = ""

        if table_name:
            if table_name and not isinstance(table_name, list):
                where_condition += f"""AND all_columns.table_name='{table_name}'\n"""
            else:
                where_condition += f"""AND all_columns.table_name IN ({list_values_in_str_with_single_quotes(
                    list_columns=table_name
                )})\n"""

        if table_schema:
            if table_schema and not isinstance(table_schema, list):
                where_condition += f"""AND all_columns.table_schema='{table_schema}'\n"""
            else:
                where_condition += f"""AND all_columns.table_schema IN ({list_values_in_str_with_single_quotes(
                    list_columns=table_schema
                )})\n"""

        sql_query = f"""
        SELECT
            all_columns.table_schema,
            schema_info.schema_description,
            all_columns.table_name,
            table_info.table_description AS table_description,
            all_columns.column_name,
            all_columns.data_type,
            columns_info.description AS column_description
        FROM
            information_schema.columns AS all_columns
        LEFT JOIN (
            SELECT
                *
            FROM
                pg_catalog.pg_statio_all_tables AS st
            LEFT JOIN pg_catalog.pg_description pgd
                ON pgd.objoid = st.relid
            LEFT JOIN information_schema.columns AS c
                ON pgd.objsubid = c.ordinal_position
                AND c.table_schema = st.schemaname
                AND c.table_name = st.relname
            ) AS columns_info
                ON all_columns.table_schema = columns_info.schemaname
                AND all_columns.table_name = columns_info.relname
                AND columns_info.column_name = all_columns.column_name
        LEFT JOIN (
            SELECT
                *,
                pg_catalog.obj_description(pgc.oid, 'pg_class') AS table_description
            FROM
                information_schema.tables AS t
            INNER JOIN pg_catalog.pg_class AS pgc
                ON t.table_name = pgc.relname
            WHERE
                t.table_type = 'BASE TABLE'
                AND pg_catalog.obj_description(pgc.oid, 'pg_class') IS NOT NULL
            ) AS table_info
                ON
                    table_info.table_name = all_columns.table_name
                    AND table_info.table_schema = all_columns.table_schema
        INNER JOIN (
            SELECT
                nspname,
                obj_description(oid) AS schema_description
            FROM
                pg_catalog.pg_namespace
            ) AS schema_info
                ON schema_info.nspname = all_columns.table_schema
        WHERE
            all_columns.table_schema != 'pg_catalog'
            {where_condition}
        """

        return self.execute_to_df(sql_query=sql_query)



from collections.abc import Sequence

import clickhouse_connect
import pandas as pd
from clickhouse_connect.driver import Client

from connectors_to_databases.BaseOperator import BaseOperator
from connectors_to_databases.TypeHinting import SQLQuery


class ClickHouse(BaseOperator):
    """Connector to ClickHouse database."""

    def __init__(
            self,
            host: str = "localhost",
            port: int = 8123,
            login: str = "default",
            password: str = "default",  # noqa: S107
    ):
        """
        Init class.

        :param host: Host/IP database; default 'localhost'.
        :param port: port database; default '8123'.
        :param login: login to database; default 'default'.
        :param password: password to database; default 'default'.
        """
        super().__init__(host, port, login, password)
        self._host = host
        self._login = login
        self._password = password
        self._port = port

    def _authorization_database(self) -> clickhouse_connect.driver.httpclient.Client:
        """Creating connector engine to database ClickHouse."""  # noqa D401

        return clickhouse_connect.get_client(
            host=self._host,
            port=self._port,
            username=self._login,
            password=self._password,
        )

    def execute_script(
            self,
            manual_sql_script: SQLQuery = None,
    ) -> None:
        """
        Execute manual scripts (INSERT, TRUNCATE, DROP, CREATE, etc.). Other than SELECT.

        :param manual_sql_script: query with manual script; default `''`.
        :return: None.
        """

        ch = self._authorization_database()

        ch.query(query=manual_sql_script)

    def execute_to_df(
            self,
            sql_query: SQLQuery = None,
    ) -> pd.DataFrame | Exception:
        """
        Getting data from database with SQL-query.

        :param sql_query; default `''`.
        :return: DataFrame with data from database.
        """  # noqa D415

        ch = self._authorization_database()

        return ch.query_df(
            query=sql_query,
        )

    def insert_df(
            self,
            df: pd.DataFrame = None,
            table_name: str = None,
            table_schema: str = None,
            dtype: Sequence[str] | None = None,
            **kwargs,
    ) -> None:
        """
        Inserting data from dataframe to database.

        Does not create a table if not exists.

        :param df: dataframe with data; default None.
        :param table_name: name of table; default None.
        :param table_schema: name of schema; default None.
        :param dtype: ClickHouse column type names; default None.
        :return: None,
        """  # noqa: D401

        ch = self._authorization_database()

        ch.insert_df(
            df=df,
            table=table_name,
            database=table_schema,
            column_type_names=dtype,
        )

    def get_uri(self) -> Client:
        """
        Get connector for manual manipulation with connect to database.

        :return engine.base.Engine:
        """

        return self._authorization_database()



from abc import ABC, abstractmethod
from urllib.parse import quote

import pandas as pd
from sqlalchemy import create_engine, engine, text

from connectors_to_databases.TypeHinting import SQLQuery


class BaseOperator(ABC):
    """BaseOperator for databases."""

    def __init__(
            self,
            host: str = "localhost",
            port: int = None,
            database: str = None,
            login: str = None,
            password: str = None,
    ):
        """
        Init class.

        :param host: Host/IP database; default 'localhost'.
        :param database: name database; default 'None'.
        :param port: port database; default 'None'.
        :param login: login to database; default 'None'.
        :param password: password to database; default 'None'.
        """
        self._host = host
        self._database = database
        self._login = login
        self._password = password
        self._port = port

    @abstractmethod
    def _authorization_database(self) -> engine.base.Engine:
        """
        Creating connector engine to some database.

        :return: engine for database.
        """ # noqa D415

        engine_str = f"base://" \
                     f"{self._login}:{quote(self._password)}@{self._host}:{self._port}/" \
                     f"{self._database}"

        return create_engine(engine_str)

    def insert_df(
            self,
            df: pd.DataFrame = None,
            table_name: str = None,
            table_schema: str = None,
            chunksize: int | None = 10024,
            index: bool = False,
            if_exists: str = "append",
            dtype: None | dict = None,
    ) -> None:
        """
        Inserting data from dataframe to database.
         
        :param df: dataframe with data; default None.
        :param table_name: name of table; default None.
        :param table_schema: name of schema; default None.
        :param chunksize: Specify the number of rows in each batch to be written at a time.
            By default, all rows will be written at once; default `10024`.
        :param index: Write DataFrame index as a column. Uses `index_label` as the column
            name in the table.
        :param if_exists: {'fail', 'replace', 'append'}, default 'append'
            How to behave if the table already exists.

            * fail: Raise a ValueError.
            * replace: Drop the table before inserting new values.
            * append: Insert new values to the existing table.
        :param dtype: Specifying the datatype for columns. If a dictionary is used, the
            keys should be the column names and the values should be the
            SQLAlchemy types or strings for the sqlite3 legacy mode. If a
            scalar is provided, it will be applied to all columns.
            
        Example:
        -------
            Create df
            >>> from connectors_to_databases import PostgreSQL
            >>> import sqlalchemy
            >>> from sqlalchemy.dialects.postgresql import UUID
            
            >>> pg = PostgreSQL()
            >>> dict_ = {'id': '41e5091e-6e97-4670-a4c9-7d6d4cc7c2af', 'date': '2020-01-01', 'amount': 100}
            >>> df_foo = pd.DataFrame([dict_]) # noqa: PD901
            >>> pg.insert_df(
            ...    df=df_foo, 
            ...    table_name='tmp_fct_sales', 
            ...    table_schema='public',
            ...    dtype={
            ...        'id': UUID,
            ...        'date': sqlalchemy.Date
            ...    }
            ... )

        """ # noqa D415

        df.to_sql(
            name=table_name,
            schema=table_schema,
            con=self._authorization_database(),
            chunksize=chunksize,
            index=index,
            if_exists=if_exists,
            dtype=dtype,
        )

    def execute_to_df(
            self,
            sql_query: SQLQuery = None,
    ) -> pd.DataFrame | Exception:
        """
        Getting data from database with SQL-query.

        :param sql_query; default `''`.
        :return: DataFrame with data from database.
        """ # noqa D415

        return pd.read_sql(
            sql=sql_query,
            con=self._authorization_database(),
        )

    def execute_script(
            self,
            manual_sql_script: SQLQuery = None,
    ) -> None:
        """
        Execute manual scripts (INSERT, TRUNCATE, DROP, CREATE, etc.). Other than SELECT.

        :param manual_sql_script: query with manual script; default `''`.
        :return: None.
        """
        with self._authorization_database().engine.begin() as conn:
            conn.execute(
                text(manual_sql_script),
            )

    def get_uri(self) -> engine.base.Engine:
        """
        Get connector for manual manipulation with connect to database.

        :return engine.base.Engine:
        """

        return self._authorization_database().engine

    def check_connection_to_database(self) -> bool | Exception:
        """
        Method to check connection to database.

        :return: boolean True, if connection to database is successful, Exception otherwise.
        """ # noqa D415
        df = self.execute_to_df("SELECT 1 AS ONE")

        return bool(isinstance(df, pd.DataFrame))



def list_values_in_str_with_double_quotes(list_columns: list = None) -> str:
    """
    **Function: list_values_in_str_with_double_quotes.**

    This function takes a list of values as the `list_columns` parameter and returns a string where each
    value is enclosed in double quotes.

    **Parameters:**
    - `list_columns` (list): The list of values to be enclosed in double quotes.
    If not specified, defaults to `None`.

    **Return:**
    - `str`: A string containing values enclosed in double quotes and separated by commas.

    **Example Usage:**

    ```python
    columns = ['column1', 'column2', 'column3']
    result = list_values_in_str_with_double_quotes(columns)
    print(result)
    ```

    **Output:**

    ```
    "column1", "column2", "column3"
    ```

    In this example, we pass the `columns` list of columns to the `list_values_in_str_with_double_quotes`
    function and store the result in the `result` variable. Then, we print the value of `result`, which will
    contain the strings from the `columns` list enclosed in double quotes and separated by commas.


    @param list_columns: The list of columns to be enclosed in double quotes. If not specified, defaults
        to `None`.; default 'None'
    @return: A string containing values enclosed in double quotes and separated by commas.
    """ # noqa D415

    return ", ".join([f'"{value}"' for value in list_columns])


def list_values_in_str_with_single_quotes(list_columns: list = None) -> str:
    """
    **Function: list_values_in_str_with_single_quotes.**

    This function takes a list of values as the `list_columns` parameter and returns a string where each
    value is enclosed in single quotes.

    **Parameters:**
    - `list_columns` (list): The list of values to be enclosed in single quotes.
    If not specified, defaults to `None`.

    **Return:**
    - `str`: A string containing values enclosed in single quotes and separated by commas.

    **Example Usage:**

    ```python
    columns = ['column1', 'column2', 'column3']
    result = list_values_in_str_with_single_quotes(columns)
    print(result)
    ```

    **Output:**

    ```
    "column1", "column2", "column3"
    ```

    In this example, we pass the `columns` list of columns to the `list_values_in_str_with_single_quotes`
    function and store the result in the `result` variable. Then, we print the value of `result`, which will
    contain the strings from the `columns` list enclosed in single quotes and separated by commas.


    @param list_columns: The list of columns to be enclosed in single quotes. If not specified, defaults
        to `None`.; default 'None'.
    @return: A string containing values enclosed in single quotes and separated by commas.
    """ # noqa D415

    return ", ".join([f"'{value}'" for value in list_columns])



from urllib.parse import quote

from sqlalchemy import create_engine, engine

from connectors_to_databases.BaseOperator import BaseOperator


class MariaDB(BaseOperator):
    """Connector to MariaDB database."""

    def __init__(
            self,
            host: str = "127.0.0.1",
            port: int = 3306,
            database: str = "sys",
            login: str = "root",
            password: str = "root",  # noqa: S107
    ):
        """
        Init class.

        :param host: Host/IP database; default 'localhost'.
        :param database: name database; default 'sys'.
        :param port: port database; default 3306.
        :param login: login to database; default 'root'.
        :param password: password to database; default 'root'.
        """
        super().__init__(host, port, database, login, password)
        self._host = host
        self._database = database
        self._login = login
        self._password = password
        self._port = port

    def _authorization_database(self) -> engine.base.Engine:
        """Creating connector engine to database MariaDB.""" # noqa D401

        # mariadb+mariadbconnector
        engine_str = f"mysql+pymysql://" \
                     f"{self._login}:{quote(self._password)}@{self._host}:{self._port}/" \
                     f"{self._database}"

        return create_engine(engine_str)
